"""
    creating code objects and keeping track of their source code 
    with artificial unique filenames. Provides and extends some 
    'inspect' functions which usually only work with code coming 
    from static modules. 
"""
from __future__ import generators
import sys, inspect
from linecache import getline as linecache_getline


def gettb(tb, n=-1):
    """ return the n'th traceback. """
    tblist = [tb]
    while tb.tb_next:
        tb = tb.tb_next
        tblist.append(tb)
    return tblist[n]


dynfileregistry = {}

def deindent(pysource):
    """ return a list of deindented lines from the given python
        source code.  The first indentation offset of a non-blank
        line determines the deindentation-offset for all the lines. 
        Subsequent lines which have a lower indentation size will
        be copied verbatim as they are assumed to be part of
        multilines. 
    """
    lines = []
    indentsize = 0
    for line in pysource.split('\n'):
        line = line.expandtabs()
        if not lines:
            if not line.strip():
                continue # skip first empty lines 
            indentsize = len(line) - len(line.lstrip())
        if line.strip():
            if len(line) - len(line.lstrip()) >= indentsize:
                line = line[indentsize:]
        lines.append(line + '\n')

    # treat trailing whitespace-containing lines correctly 
    # (the python parser at least in python 2.2. is picky about it)
    while lines:
        line = lines.pop()
        if not line.strip():
            continue
        if not line.endswith('\n'):
            line = line + '\n'
        lines.append(line)
        break
    return lines

def makecode(source, mode='exec', flag=generators.compiler_flag):
    lines = deindent(source)
    source = "".join(lines)
    frame = inspect.currentframe(1) # the caller
    origin = "%s:%d" % (frame.f_code.co_filename, 
                        frame.f_lineno)
    filename = makedynfilename(origin, lines)
    try:
        #print "compiling source:", repr(source)
        co = compile(source, filename, mode, flag)
    except SyntaxError, ex:
        # re-represent syntax errors from parsing python strings 
        newex = SyntaxError('\n'.join([
            "".join(lines[:ex.lineno]),
            " " * ex.offset + '^', 
            "syntax error probably generated here: %s" % origin]))
        newex.offset = ex.offset
        newex.lineno = ex.lineno
        newex.text = ex.text
        raise newex

    dynfileregistry[filename] = origin, lines, co 
    return co

def installhacks():
    # a hack for convenience of displaying tracebacks 
    import linecache
    linecache.getline = getline

def tbinfo(tb, index = -1):
    """ return an (filename:lineno, linecontent) tuple for the given 
        traceback.  Works also with code generated from makecode(). 
    """
    tb = gettb(tb, index)
    filename = tb.tb_frame.f_code.co_filename
    lineno = tb.tb_lineno 
    return filename + ':%d' % lineno, getline(filename, lineno)

def makedynfilename(origin, source, originmap={}):
    entry = originmap.setdefault(origin, [])
    num = len(entry)
    entry.append(num)
    if num > 0:
        return "<%s [%d]>" % (origin, num)
    else:
        return "<%s>" % origin 

def getline(filename, lineno):
    """ return line in filename (possibly dyncode). """
    tup = dynfileregistry.get(filename, None)
    if tup is not None:
        origin, lines, co = tup
        return lines[lineno-1]
    else:
        return linecache_getline(filename, lineno)

def getsource(obj):
    try:
        inspect.getsource(obj)
    except IOError:
        for attr in 'func_code',:
            code = getattr(obj, attr, None)
            if code is not None:
                filename = code.co_filename
                if dynfileregistry.has_key(filename):
                    break
                raise IOError, "could not get source (even after looking in the code registry)"
        else:
            raise

        origin, lines, gencode = dynfileregistry[filename]
        firstlineno = code.co_firstlineno - 1
        lines = inspect.getblock(lines[firstlineno:])
        return "".join(lines)

def getpyfile(obj):
    fn = inspect.getfile(obj)
    if fn.endswith('.pyc'):
        fn = fn[:-1]
    assert fn.endswith('.py')
    return fn
