" File: funclist.vim
" Author: Yegappan Lakshmanan
" Version: 1.3
" Last Modified: May 8, 2002
"
" Overview
" --------
" The "Function List" (funclist.vim) plugin script opens a Vim window with
" a list of functions defined in the current file. When you select a function
" name from this window, the cursor will be positioned at the definition
" of the function in the source file. This script will run in both GUI
" and console/terminal version of Vim.
"
" This script relies on the exuberant ctags utility
" (http://ctags.sourceforge.net) to generate the function listing.  This
" script will run on all the platforms where the exuberant ctags utility is
" supported (this includes MS-Windows and Unix based systems).
"
" This script supports the following language files: Assembly, ASP, Awk, C,
" C++, Cobol, Eiffel, Fortran, Java, Lisp, Make, Pascal, Perl, PHP, Python,
" Rexx, Ruby, Scheme, Shell, Slang, TCL and Vim.
"
" You can select a function either by pressing the <Enter> key or by double
" clicking the name using a mouse.
"
" The script will automatically highlight the name of the current function.
" The function name will be highlighted after 'updatetime' milliseconds. The
" default value for this Vim option is 4 seconds. The highlighting will work
" only if the function list is sorted by order. If the function listing is
" sorted by name, the function highlighting will not work.
"
" If you place the cursor on a function name in the "Function List" window,
" then the function prototype will be displayed at the Vim status line after
" 'updatetime' milliseconds. The default value for the 'updatetime' Vim option
" is 4 seconds.
"
" This script relies on the Vim "filetype" detection mechanism to determine
" the type of the current file. To turn on filetype detection use
"
"               :filetype on
"
" This script will not work in 'compatible' mode.  Make sure the 'compatible'
" option is not set.
"
" Configuration
" -------------
" By changing the following variables you can configure the behavior of this
" script. Set the following variables in your .vimrc file using the 'let'
" command.
"
" The script uses the Flist_Ctags_Cmd variable to locate the ctags utility.
" By default, this is set to /usr/bin/ctags. Set this variable in your .vimrc
" file to point to the location of the ctags utility in your system
"
"               let Flist_Ctags_Cmd = 'd:\tools\ctags.exe'
"
" To open the function list window, you have to press the key defined by the
" Flist_Key variable. By default, this variable is set to the <F8> key.
" You can modify this to a different key in your .vimrc file:
"
"               let Flist_Key = '\l'
"
" When the cursor is not moved for a period specified by the 'updatetime'
" Vim option, the script will automatically highlight the function under the
" cursor. To force the current function name highlight, you can press the key
" defined by the Flist_Sync_Key variable. By default, this variable is set
" to the <F9> key. You can modify this to a different key in your .vimrc file:
"
"               let Flist_Sync_Key = '\h'
"
" By default, the function names will be listed in the order in
" which they are defined in the file. You can alphabetically sort the function
" names by pressing the "s" key in the function list window. You can also
" change the default order by setting the variable Flist_Sort_Type to
" "name" or "order" in your .vimrc file:
"
"               let Flist_Sort_Type = "name"
"
" Be default, the function names will be listed in a vertically split window.
" If you prefer a horizontally split window, then set the
" 'Flist_Use_Horiz_Window' variable to 1 in your .vimrc file:
"
"               let Flist_Use_Horiz_Window = 1
"
" By default, the function names will be listed in only one window. The
" window will be reused for listing functions from different files. If you
" prefer to open a function list window for each file separately then set the
" 'Flist_Use_One_Window' variable to 0:
"
"               let Flist_Use_One_Window = 0
"
" To automatically refresh the function list window as you switch between
" buffers, set the Flist_Auto_Refresh variable to 1. By default, this variable
" is set to 0 and the automatic refresh is disabled.
"
"               let Flist_Auto_Refresh = 1
"
" By default, only the function name will be displayed in the function list
" window. If you like to see function prototypes instead of names, set the
" Flist_Display_Prototype variable to 1. By default, this variable is set to 0
" and function names will be displayed.
"
"               let Flist_Display_Prototype = 1
"
" By default, when the width of the window is less than 100 and a new function
" list window is opened vertically, then the window width will be increased by
" 20 to accomodate for the new window. If you have any screen refresh problems
" or if you prefer not to adjust the window size then set the
" 'Flist_Inc_Winwidth' variable to 0:
"
"               let Flist_Inc_Winwidth = 0
"
" The default width of vertically split function list window will be 20.
" This can be changed by modifying the Flist_WinWidth variable:
"
"               let Flist_WinWidth = 30
"
" TODO:
" 1. Display variables, classes, structures, etc.
" 2. Group the different types of tags (variables, functions, etc)

" ****************** Do not modify after this line ************************

if exists("loaded_funclist") || &cp
    finish
endif
let loaded_funclist=1

" The default location of the exuberant ctags
if !exists("Flist_Ctags_Cmd")
    "let Flist_Ctags_Cmd = '/usr/bin/ctags'
    let Flist_Ctags_Cmd = 'd:\vim\ctags.exe'
endif

" Key to open the function listing window
if !exists("Flist_Key")
    let Flist_Key = '<F8>'
endif

" Key to highlight the current function
if !exists("Flist_Sync_Key")
    let Flist_Sync_Key = '<F9>'
endif

" Function listing sort type
if !exists("Flist_Sort_Type")
    let Flist_Sort_Type = "order"
endif

" Use horizontally split window
if !exists("Flist_Use_Horiz_Window")
    let Flist_Use_Horiz_Window = 0
endif

" Increase the window width to show the function list window
if !exists("Flist_Inc_Winwidth")
    let Flist_Inc_Winwidth = 1
endif

" Use only one window for listing the functions in all the files or use
" multiple windows
if !exists("Flist_Use_One_Window")
    let Flist_Use_One_Window = 1
endif

" Width of the vertically split function list window
if !exists("Flist_WinWidth")
    let Flist_WinWidth = 20
endif

" Auto refresh the function display window
if !exists("Flist_Auto_Refresh")
    let Flist_Auto_Refresh = 0
endif

" Display function prototypes or function names
if !exists("Flist_Display_Prototype")
    let Flist_Display_Prototype = 0
endif

" Map the key to open the function window
exe "nnoremap <unique> <silent> " . Flist_Key . 
            \ " :call <SID>FlistToggleWindow(expand('%:p'), &filetype)<CR>"
exe "nnoremap <unique> <silent> " . Flist_Sync_Key . 
            \ " :call <SID>Flist_HiLite_Func_Name()<CR>"

" Colors used to highlight the selected function name
highlight clear FunctionName
if has("gui_running") || &t_Co> 2
    highlight FunctionName term=reverse cterm=bold
    highlight FunctionName ctermfg=0 ctermbg=3 guifg=Black guibg=Yellow
else
    highlight FunctionName term=reverse cterm=reverse
endif

function! s:FlistToggleWindow(filename, ftype)
    " Function list window name
    if g:Flist_Use_One_Window == 1
        let winname = "__Function_List__"
    else
        let winname = "__" . a:filename . "__Function_List__"
    endif

    " If window is open then close it Close the window only if the current the
    " function listing is for the current file
    let winnum = bufwinnr(winname)
    if winnum != -1 && getwinvar(winnum, 'filename') == expand("%:p")
        let curbufnr = bufnr("%")
        exe winnum . "wincmd w"
        close
        let winnum = bufwinnr(curbufnr)
        if winnr() != winnum
            exe winnum . "wincmd w"
        endif
        return
    endif

    " Open the function list window
    call s:FlistOpenWindow(a:filename, a:ftype)
endfunction

" FlistOpenWindow
" List the functions defined in the specified file in a Vim window
function! s:FlistOpenWindow(filename, ftype)
    " Make sure the file name is not empty and is readable
    if a:filename == "" || !filereadable(a:filename)
        return
    endif

    " Exuberant ctags arguments to generate a function list
    let ctags_args = " --format=2 --fields=n -f - "

    " Get the sort type. First time, use the global setting. After that use
    " the window local variable
    if !exists("w:sort_type")
        let sort_type = g:Flist_Sort_Type
    else
        let sort_type = w:sort_type
    endif

    " Form the ctags argument depending on the sort type 
    if sort_type == "name"
        let ctags_args = ctags_args . " --sort=yes "
    else
        let ctags_args = ctags_args . " --sort=no "
    endif

    " Specify ctags argument depending on the filetype
    if a:ftype == 'asm'
        let ft_args = " --language-force=asm --asm-types=l "
    elseif a:ftype == 'aspperl' || a:ftype == 'aspvbs'
        let ft_args = " --language-force=asp --asp-types=fs "
    elseif a:ftype == 'awk'
        let ft_args = " --language-force=awk --awk-types=f "
    elseif a:ftype == 'c'
        let ft_args = " --language-force=c --c-types=f "
    elseif a:ftype == 'cpp'
        let ft_args = " --language-force=c++ --c++-types=f "
    elseif a:ftype == 'cobol'
        let ft_args = " --language-force=cobol --cobol-types=p "
    elseif a:ftype == 'eiffel'
        let ft_args = " --language-force=eiffel --eiffel-types=f "
    elseif a:ftype == 'fortran'
        let ft_args = " --language-force=fortran --fortran-types=f "
    elseif a:ftype == 'java'
        let ft_args = " --language-force=java --java-types=m "
    elseif a:ftype == 'lisp'
        let ft_args = " --language-force=lisp --lisp-types=f "
    elseif a:ftype == 'make'
        let ft_args = " --language-force=make --make-types=m "
    elseif a:ftype == 'pascal'
        let ft_args = " --language-force=pascal --pascal-types=f "
    elseif a:ftype == 'perl'
        let ft_args = " --language-force=perl --perl-types=s "
    elseif a:ftype == 'php'
        let ft_args = " --language-force=php --php-types=f "
    elseif a:ftype == 'python'
        let ft_args = " --language-force=python --python-types=f "
    elseif a:ftype == 'rexx'
        let ft_args = " --language-force=rexx --rexx-types=s "
    elseif a:ftype == 'ruby'
        let ft_args = " --language-force=ruby --ruby-types=f "
    elseif a:ftype == 'scheme'
        let ft_args = " --language-force=scheme --scheme-types=f "
    elseif a:ftype =~ '\<[cz]\=sh\>'
        let ft_args = " --language-force=sh --sh-types=f "
    elseif a:ftype == 'slang'
        let ft_args = " --language-force=slang --slang-types=f "
    elseif a:ftype == 'tcl'
        let ft_args = " --language-force=tcl --tcl-types=p "
    elseif a:ftype == 'vim'
        let ft_args = " --language-force=vim --vim-types=f "
    else
        echohl WarningMsg
        echon "Function listing not supported for " . a:ftype . " filetype"
        echohl None
        return
    endif

    " Ctags command to produce output with regexp for locating the functions
    let ctags_cmd = g:Flist_Ctags_Cmd . ctags_args . ft_args
    let ctags_cmd = ctags_cmd . '"' . a:filename . '"'

    " Get the function list
    let ctags_cmd_output = system(ctags_cmd)

    if v:shell_error && ctags_cmd_output != ""
        echohl WarningMsg | echon ctags_cmd_output | echohl None
        return
    endif

    if ctags_cmd_output == ""
        echohl WarningMsg
        echomsg "No functions found for " . a:filename
        echohl None
        return
    endif

    " Function list window name
    if g:Flist_Use_One_Window == 1
        let winname = "__Function_List__"
    else
        let winname = "__" . a:filename . "__Function_List__"
    endif

    let winnum = bufwinnr(winname)
    if winnum == -1
        " Create a new window. If user prefers a horizontal window, then
        " open a horizontally split window. Otherwise open a vertically
        " split window
        if g:Flist_Use_Horiz_Window == 1
            if g:Flist_Use_One_Window == 1
                botright 10new
            else
                rightbelow 10new
            endif
        else
            " Increase the window size, if needed, to accomodate the new
            " window
            if g:Flist_Inc_Winwidth == 1 && 
                        \ &columns < (80 + g:Flist_WinWidth)
                " one extra column is needed to include the vertical split
                let &columns= &columns + (g:Flist_WinWidth + 1)
            endif

            " If only one function list window should be used, then open the
            " window at the leftmost place. Otherwise open a new vertically
            " split window
            if g:Flist_Use_One_Window == 1
                exe "topleft " . g:Flist_WinWidth . "vnew"
            else
                exe g:Flist_WinWidth . "vnew"
            endif
        endif
        let w:FlistFuncCount = 0

        let bufnum = bufnr(winname)
        if bufnum == -1
            " Create a new buffer
            exe "edit " . winname
        else
            " Edit the existing buffer
            exe "buffer " . winname
        endif

        " Mark buffer as scratch
        silent! setlocal buftype=nofile
        silent! setlocal bufhidden=delete
        silent! setlocal noswapfile
        silent! setlocal nowrap
        silent! setlocal buflisted
    else
       if winnr() != winnum
            " Jump to the existing window
            exe winnum . "wincmd w"
        endif
        " Make the buffer modifiable
        silent! setlocal modifiable
        " Goto the first line in the buffer
        go
        " Delete the contents of the buffer to the black-hole register
        %delete _
    endif

    " Clear the function name highlight
    match none

    call append(0, '" Function listing for ' . a:filename)
    call append(1, '" Press s to sort by name/order')

    " Clear the function name highlight
    match none

    let w:Flist_comment_lines = 3

    if g:Flist_Display_Prototype == 0
        " Extract all the function names from the ctags output
        let pat = "\\([^\t]\\+\\)[^\n]\\+\n"
        let tagnames = substitute(ctags_cmd_output, pat, "\\1\n", "g")

        " Add the function names to the buffer
        silent! put =tagnames
    else
        " Extract all the function names from the ctags output
        let pat = "[^\t]\\+\t[^\t]\\+\t\/^\\(.\\{-}\\)\\\$\/;[^\n]\\+\n"
        let prototypes = substitute(ctags_cmd_output, pat, "\\1\n", "g")

        " Add the function names to the buffer
        silent! put =prototypes
    endif

    " Store the ctags output and the filename in window local variables
    let w:filename = a:filename
    let w:ftype = a:ftype
    let w:sort_type = sort_type

    " Goto the first line in the buffer
    go

    " Create a mapping for jumping to the functions and sorting the list
    nnoremap <buffer> <silent> <CR> :call <SID>FlistJumpFunction()<CR>
    nnoremap <buffer> <silent> <2-LeftMouse> :call <SID>FlistJumpFunction()<CR>
    nnoremap <buffer> <silent> s :call <SID>FlistChangeSort()<CR>

    " Highlight the comments
    if has("syntax") && !has("syntax_items")
        syn match FunctionListSynopsis    "^\" .*"
        hi def link FunctionListSynopsis    Special
    endif

    " Mark the buffer as not modifiable
    silent! setlocal nomodifiable

    " Create window local variables, one for each function. The format
    " of the variable is w:Flist{function_index}. Each variable will be
    " set to the corresponding function search pattern. This will be used
    " to display function prototypes and to jump to the function.
    " The window local w:FlistFuncCount variable contains the number of
    " functions currently displayed in the window

    " First delete all the old variables
    let i = w:FlistFuncCount
    while i > w:Flist_comment_lines
        unlet! w:Flist{i}
        unlet! w:Flist_lnum{i}
        let i = i - 1
    endwhile

    " We start the function count from n instead of 1. This is to account for
    " the comment lines displayed at the top of the window and to match the
    " function index with the function line number in the window for easy
    " lookup

    " Separate all the ctags output lines and extract the line numbers and
    " store them
    let i = w:Flist_comment_lines + 1

    let len = strlen(ctags_cmd_output)
    while ctags_cmd_output != ""
        let one_line = strpart(ctags_cmd_output, 0,
                    \ stridx(ctags_cmd_output, "\n"))
        let mx = "line:\\zs\\d\\+\\ze$"
        let w:Flist_lnum{i} = matchstr(one_line, mx)
        let w:Flist{i} = one_line

        let i = i + 1
        let ctags_cmd_output = strpart(ctags_cmd_output, 
                    \ stridx(ctags_cmd_output, "\n") + 1, len)
    endwhile

    let w:FlistFuncCount = i - 1

    if g:Flist_Auto_Refresh == 1
        wincmd p
    endif
endfunction

function! s:FlistChangeSort()
    if !exists("w:filename") || !exists("w:ftype")
        return
    endif

    if w:sort_type == "name"
        let w:sort_type = "order"
    else
        let w:sort_type = "name"
    endif

    call s:FlistOpenWindow(w:filename, w:ftype)
endfunction

" FlistJumpFunction
" Jump to the location of the selected function definition
function! s:FlistJumpFunction()
    " Do not process comment lines and empty lines
    let curline = getline(".")
    if curline == "" || curline[0] == '"'
        return
    endif

    if g:Flist_Display_Prototype == 0
        " Word under cursor
        let funcname = curline
    endif

    " Get the tag output for the current function
    let mtxt = w:Flist{line(".")}

    if mtxt == ""
        return
    endif

    " Extract the tagname, file name and the search pattern
    let mx = "\\([^\t]\\+\\)\t\\([^\t]\\+\\)\t\/^\\(.*\\)\\\$\/;.*"
    let tagname = substitute(mtxt, mx, '\1', '')
    let filename = substitute(mtxt, mx, '\2', '')
    let tagpat =  substitute(mtxt, mx, '\\\V\\\^\3\\\$', '')

    " Make sure the tag name and the file name are correct
    if g:Flist_Display_Prototype == 0
        if tagname != funcname || filename != w:filename
            return
        endif
    else
        if filename != w:filename
            return
        endif
    endif

    " Clear previously selected name
    match none
    " Highlight the current selected name
    exe "match FunctionName /\\%" . line(".") . "l.*/"

    " Goto the window containing the file.  If the window is not there,
    " open a new window
    let winnum = bufwinnr(w:filename)
    if winnum == -1
        let bufnum = bufnr(w:filename)
        if g:Flist_Use_Horiz_Window == 1
            exe "leftabove split #" . bufnum
        else
            exe "rightbelow vertical split #" . bufnum
            wincmd p
            exe "vertical resize " . g:Flist_WinWidth
            wincmd p
        endif
    else
        exe winnum . "wincmd w"
    endif

    " Jump to the function
    silent call search(tagpat, 'w')

    " Bring the line to the middle of the window
    normal! z.
endfunction

" Display the prototype of the function under the cursor
function! s:Flist_Show_Prototype()
    " If we have already display prototype in the function window, no need
    " to display it in the status line
    if g:Flist_Display_Prototype == 1
        return
    endif

    " Clear the previously displayed line
    echo

    " Do not process comment lines and empty lines
    let curline = getline(".")
    if curline == "" || curline[0] == '"'
        return
    endif

    " Get the tag output for the current function
    let mtxt = w:Flist{line(".")}

    if mtxt == ""
        return
    endif

    " Get the function search pattern and display it
    let mx = "\t\/^\\zs.*\\ze\\\$\/;.*$"

    echo matchstr(mtxt, mx)
endfunction

" This function does binary search in the array of tag names and highlights
" the corresponding function name. The idea behind this function is taken
" from the ctags.vim script available at Vim online website.
function! s:Flist_HiLite_Func_Name()

    " Get the current file name
    let cur_fname = expand("%:p")
    if cur_fname == ""
        return
    endif

    " Form the name of the function listing window
    if g:Flist_Use_One_Window == 1
        let winname = "__Function_List__"
    else
        let winname = "__" . cur_fname . "__Function_List__"
    endif

    " Make sure the function listing window is present
    let winnum = bufwinnr(winname)
    if winnum == -1
        return
    endif

    " Make sure we have the function listing for the current file
    if getwinvar(winnum, 'filename') != cur_fname
        return
    endif

    " If the function listing is sorted by name, do not highlight.
    " We need the function listing to be sorted by order
    if getwinvar(winnum, 'sort_type') == 'name'
        return
    endif

    " Save the original window number
    let org_winnr = winnr()
    let curline = line(".")

    " Goto the function listing window
    exe winnum . "wincmd w"

    " Clear previously selected name
    match none

    let left = w:Flist_comment_lines + 1
    let right = w:FlistFuncCount

    " If cursor is behind the first function, then nothing needs
    " to be highlighted
    if curline < w:Flist_lnum{left}
        exe org_winnr . "wincmd w"
        return
    endif

    " Do a binary search for the function name to highlight based on the line
    " numbers
    while left < right
        let middle = (right + left + 1) / 2
        let middleline = w:Flist_lnum{middle}

        if middleline == curline
            let left = middle
            break
        endif

        if middleline > curline
            let right = middle - 1
        else
            let left = middle
        endif
    endwhile

    " Bring the function name to the middle of the display
    exe left
    normal! z.

    " Highlight the function name
    exe "match FunctionName /\\%" . left . "l.*/"

    " Go back to the original window
    exe org_winnr . "wincmd w"

    return
endfunction

function! s:Flist_Close_Window()
    if g:Flist_Use_Horiz_Window || 
                \ &columns < (80 + g:Flist_WinWidth) || 
                \ g:Flist_Inc_Winwidth == 0
        " No need to adjust window width if horizontally split function
        " listing window or if columns is less than 101 or if the user chose
        " not to adjust the window width
        return
    endif

    if g:Flist_Use_One_Window == 1
        " Only one window is used for listing functions and that window is
        " getting closed, so adjust the width
        let &columns= &columns - (g:Flist_WinWidth + 1)
    else
        " Multiple windows are used for listing functions, adjust only if
        " needed
        let i = 1
        let cnt = 0
        let bufno = winbufnr(i)

        " Make sure no function listing window is open
        while bufno != -1
            if bufname(bufno) =~ "__Function_List__"
                let cnt = cnt + 1
            endif
            let i = i + 1
            let bufno = winbufnr(i)
        endwhile

        if cnt > 1
            " Some function listing window is still open
            return
        endif

        let &columns= &columns - (g:Flist_WinWidth + 1)
    endif
endfunction

function! s:Flist_Refresh_Window()
    " Function list window will be refreshed only for single window option
    if g:Flist_Auto_Refresh == 0 || g:Flist_Use_One_Window == 0
        return
    endif

    let fname = expand('%:p')
    let ftype = &filetype

    " Function list window name
    if g:Flist_Use_One_Window == 1
        let winname = "__Function_List__"
    else
        let winname = "__" . a:filename . "__Function_List__"
    endif

    let winnum = bufwinnr(winname)
    if winnum == -1
        " Window is not open. No need to refresh it
        return
    endif

    " If the function listing for the current window is already present, no
    " need to refresh it
    if getwinvar(winnum, 'filename') == fname
        return
    endif

    " Refresh the function list window
    call s:FlistOpenWindow(fname, ftype)
endfunction

augroup FunctionListAutoCmds
    autocmd!
    " Autocommand to display the function prototype for the function under
    " the cursor.
    autocmd CursorHold *__Function_List__ call s:Flist_Show_Prototype()
    " Autocommand to highlight the current function 
    autocmd CursorHold * silent call <SID>Flist_HiLite_Func_Name()
    autocmd BufDelete *__Function_List__ call <SID>Flist_Close_Window()
    autocmd BufEnter * call <SID>Flist_Refresh_Window()
augroup end
