"""Vim debugger for Python scripts

    Adds the following commands to Vim:

        :Vdb - start debugging
        :[line]Vbp - set a breakpoint, line defaults to the current
        :[line]Vbd - delete a breakpoint, line defaults to the current

    When you call the Vdb command, Vdb will ask you for a command to run. You
    should enter a valid Python expression there, the expression will be 
    validated and Vdb will (when not interrupted) walk you through the file 
    line by line and call by call. During this step processing it will show a
    prompt, when you just press enter Vdb will continue with the next step,
    but you can also enter a single line of Python code that will be evaluated
    or a Vdb command. When you enter a line of Python code, a simple 
    interpreter will execute the command (mind that it does *not* print the
    result of your command, so you should use the 'print' statement to see any
    results). The commands are simple debugger commands, mainly to change the
    size of the steps taken.
    Those commands are:

        s (or step): set the step size to a single call (default)
        n (or next): make the debugger only show steps in the current function
        c (or continue): only stop on breakpoints
        h (or help, or ?): show the list of commands
        q (or quit): stop debugging

    Any other input will be executed as Python code, no input will make the
    debugger continue in the current stepsize.

"""

import vim
import sys
import bdb
import inspect
import time

# dyncode module, stolen from shpy (by Holger Krekel & Armin Rigo, see 
#   http://codespeak.net), improves traceback for non-fs code
import dyncode

def escape_string(s):
    """escape all chars that can break a double-quoted string in vim"""
    s = str(s)
    # XXX guess there are more of these, right?
    s = s.replace('\\', '\\\\')
    s = s.replace('\n', '\\n')
    s = s.replace('"', '\\"')
    s = s.replace("'", "\\'")
    s = s.replace('<Esc>', '\<Esc>')
    
    return s

def import_buffer(context={}):
    datalines = vim.current.buffer[:]
    data = '\n'.join(datalines)
    code = dyncode.makecode('%s\n' % data.replace('\r', ''), 'exec')
    context.update(globals())
    context['__file__'] = code.co_filename
    exec code in context
    return context
    
def reload_buffer_module():
    bufname = vim.current.buffer.name
    if not bufname.endswith('.py'):
        return
    for mod, code in sys.modules.items():
        if hasattr(code, '__file__') and code.__file__.startswith(bufname):
            try:
                __import__(mod)
            except ImportError:
                pass
    
def request_input(p=">> ", evaluate=1):
    command = "input('%s')" % p
    vim.command('call inputsave()')
    inp = vim.eval(command)
    vim.command('call inputrestore()')
    if evaluate:
        inp = eval(inp)
    return inp

def set_breakpoint(line):
    vim.command('sign define breakpoint text=!! linehl=none texthl=none')
    line = str(line)
    breakpoints = get_breakpoints()
    signs = [i for i in vim.eval('b:vdb_breakpoints').split(' ') if i.strip()]
    newindex = 100
    if signs:
        signs.sort()
        newindex = int(signs[-1]) + 1
    if line not in breakpoints:
        signs.append(str(newindex))
        vim.command('let b:vdb_breakpoints = "%s"' % ' '.join(signs))
        vim.command('sign place %s name=breakpoint line=%s buffer=%s' % 
                        (newindex, line, vim.current.buffer.number))

def del_breakpoint(line):
    curline = vim.current.window.cursor[0]
    signs = [i for i in vim.eval('b:vdb_breakpoints').split(' ') if i.strip()]
    id = None
    for s in signs:
        vim.command('sign jump %s buffer=%s' % (s, vim.current.buffer.number))
        signline = vim.current.window.cursor[0]
        if signline == line:
            vim.command('sign unplace %s' % s)
            signs.remove(s)
            vim.command('let b:vdb_breakpoints = "%s"' % ' '.join(signs))
            break
    vim.command(str(line))

def get_breakpoints():
    # this is really nasty: to find on which linenumbers the signs are,
    # one has to go to each sign and ask the current linenumber
    vim.command("if ! exists(\"b:vdb_breakpoints\") | let b:vdb_breakpoints=\"\" | endif")
    breakpoints = [i for i in vim.eval('b:vdb_breakpoints').split(' ') if i.strip()]
    signs = []
    curline = vim.current.window.cursor[0]
    for bp in breakpoints:
        vim.command('sign jump %s buffer=%s' % (bp, vim.current.buffer.number))
        signs.append(vim.current.window.cursor[0])
    # jump back to the original position
    vim.command('%s' % curline)
    vim.command('redr')
    return signs

class StringFifo:
    """Simple string FIFO
    
        easier to use than StringIO since this seek doesn't have to 
        be called for each action
    """
    def __init__(self):
        self._b = []

    def read(self, l=None):
        s = ''.join(self._b)
        if l is not None:
            self._b = [s[l:]]
            return s[:l]
        else:
            self._b = []
            return s

    def write(self, data):
        self._b.append(data)

class VimDb(bdb.Bdb):
    """Simple debugger for Vim

        See the bdb and pdb modules for documentation
    """

    prompt = "vdb >>> "
    
    def __init__(self, modname, breakpoints=[]):
        self._modname = modname
        self.__old_stdout = sys.stdout
        self.__buffer = StringFifo()
        self._sign_set = 0
        vim.command('sign define vimdbhl text=>> linehl=Visual')
        # set an empty sign so that the indention will stay if we
        # temporarily remove the debug sign
        vim.command('sign define empty')
        vim.command('sign place 9999 name=empty line=1 buffer=%s' % 
                        self.get_buffer_index())
        sys.stdout = self.__buffer
        bdb.Bdb.__init__(self)
        for bp in breakpoints:
            self.set_break(self._modname, bp)

    def user_call(self, frame, argument_list):
        if not self.is_buffer_module(frame):
            return
        
        framerec = inspect.getframeinfo(frame)
        self.highlight_line(framerec[1])
        self.handle_output()
        args = []
        for k, v in frame.f_locals.items():
            args.append('%s: %s' % (k, v))
        vim.eval('input("call %s(%s)")' % 
                    (escape_string(framerec[2]), escape_string(', '.join(args))))
        self.process_command(request_input(self.prompt, 0), frame)

    def user_line(self, frame):
        if not self.is_buffer_module(frame):
            return
        
        framerec = inspect.getframeinfo(frame)
        self.handle_output()
        self.highlight_line(framerec[1])
        self.process_command(request_input(self.prompt, 0), frame)

    def user_return(self, frame, return_value):
        if not self.is_buffer_module(frame):
            return
        
        framerec = inspect.getframeinfo(frame)
        self.highlight_line(framerec[1])
        self.handle_output()
        vim.eval('input("%s returned %s")' % 
                    (escape_string(framerec[2]), escape_string(return_value)))
        self.process_command(request_input(self.prompt, 0), frame)

    def user_exception(self, frame, (exc, e, tb)):
        if not self.is_buffer_module(frame):
            return
        
        framerec = inspect.getframeinfo(frame)
        filename = frame.f_code.co_filename
        self.highlight_line(framerec[1])
        self.handle_output()
        vim.eval('input("exception %s %s")' %
                    (escape_string(exc), escape_string(e)))
        del tb
        self.process_command(request_input(self.prompt, 0), frame)

    def process_command(self, command, frame):
        """process a command while debugging"""
        #self.restore_line()
        command = command.strip()
        if not command:
            return
        elif command == 's' or command == 'step':
            self.set_step()
        elif command == 'n' or command == 'next':
            self.set_next(frame)
        elif command == 'c' or command == 'continue':
            self.set_continue()
        elif command == 'h' or command == 'help' or command == '?':
            self.print_help()
            vim.eval('input("")')
        elif command == 'q' or command == 'quit':
            self.set_quit()
        else:
            self.interpreter_loop(command, frame)

    def interpreter_loop(self, command, frame):
        """simple interpreter loop"""
        # if no command is given, return to the debugger loop
        while command != 'q' and command != 'quit' and command.strip():
            try:
                exec command in frame.f_locals
            except:
                exc, e, tb = sys.exc_info()
                vim.command('echo "exception %s - %s"' % 
                        (escape_string(exc), escape_string(e)))
                del tb
            self.handle_output()
            command = request_input('>>> ', 0)

    def handle_output(self):
        buffered = self.__buffer.read().strip()
        if buffered:
            sep = ' '
            if buffered.find('\n') > -1 and buffered.rfind('\n') < len(buffered) - 1:
                sep = '\\n'
            vim.eval('input("STDOUT:%s%s")' % (sep, escape_string(buffered)))
        vim.command('redr')
                
    def highlight_line(self, lineno):
        self.restore_line()
        vim.command('sign place 10000 name=vimdbhl line=%s buffer=%s' % (lineno, self.get_buffer_index()))
        vim.command(str(lineno))
        vim.command('normal %s' % lineno)
        vim.command('redr')
        self._sign_set = 1
        return

    def restore_line(self):
        if self._sign_set:
            vim.command('sign unplace 10000')
            vim.command('redr')
            self._sign_set = 0

    def get_buffer_index(self):
        """get the index of the current buffer"""
        return vim.current.buffer.number

    def is_buffer_module(self, frame):
        filename = frame.f_code.co_filename
        return (filename == self._modname or
                    filename == vim.current.buffer.name)

    def print_help(self):
        vim.command('echo " "')
        vim.command('echo "Python debugger (c) Guido Wesdorp 2004"')
        vim.command('echo " "')
        vim.command('echo "Commands:"')
        vim.command('echo "- s[tep]: step to next line"')
        vim.command('echo "- n[ext]: next in frame"')
        vim.command('echo "- c[ontinue]: continue to next breakpoint"')
        vim.command('echo "- q[uit]: quit debugging"')
        vim.command('echo " "')
        vim.command('echo "Also you can enter single-line Python statements,"')
        vim.command('echo "those will be executed in the context of the code"')
        vim.command('echo "currently being debugged."')

    def quit(self):
        self.restore_line()
        vim.command('sign unplace 9999')
        self.clear_all_breaks()
        sys.stdout = self.__old_stdout

def py_debugger():
    vim.command('normal zR')
    reload_buffer_module()
    context = import_buffer()
    modname = context['__file__']
    breakpoints = get_breakpoints()
    command = request_input('Enter command to debug: ', 0)
    db = VimDb(modname, breakpoints)
    try:
        try:
            db.run(command, context)
        except:
            exc, e, tb = sys.exc_info()
            vim.command('input("debugger exception %s (%s), quitting"' % (exc, e))
            del tb
    finally:
        db.quit()

def register_commands():
    """register the basic set of vim commands 
        
        such as setting and removing the breakpoints and running 
        the debugger
    """
    try:
        vim.command('command -range Vbp :py set_breakpoint(<line1>)')
        vim.command('command -range Vbd :py del_breakpoint(<line1>)')
        #vim.command('command Vda :py del_all_breakpoints()')
        vim.command('command Vdb :py py_debugger()')
    except vim.error:
        pass

register_commands()
